progresscell progress {
 background: var(--accent-bg-color);
 border-right: none;
}

progresscell {
 min-height: 18px;
 border-radius: 3px;
}

progresscell trough {
 background: color-mix(in srgb, var(--border-color), transparent);
}

progresscell label {
 font-size: 0.9em;
 padding: 1px 3px;
 font-feature-settings: 'tnum';
}

progresscell label.in-progress {
 color: var(--accent-fg-color);
}

callgraphview {
}

callgraphview treeexpander indent { -gtk-icon-size: 8px; }
callgraphview treeexpander indent:nth-child(n+20) { -gtk-icon-size: 0px; }

callgraphview symbol {
  margin: -1px -3px;
  padding: 1px 6px;
}
callgraphview treeexpander expander:not(.expandable) {
  -gtk-icon-source: -gtk-icontheme("empty-symbolic");
}
callgraphview row:not(:selected) treeexpander expander:checked+box symbol.process,
callgraphview row:not(:selected) treeexpander expander:checked+box symbol.thread,
callgraphview row:not(:selected) treeexpander expander:checked+box symbol.root {
  border-radius: 9999px;
  background-color: color-mix(in srgb, currentColor 5%, transparent);
}
callgraphview row:not(:selected) treeexpander symbol.context-switch {
  border-radius: 9999px;
  background-color: color-mix(in srgb, var(--warning-color) 10%, transparent);
}
callgraphview row:not(:selected) treeexpander symbol.unwindable {
  border-radius: 9999px;
  background-color: color-mix(in srgb, var(--error-color) 10%, transparent);
}

timeruler {
  min-height: 24px;
  font-size: 0.833em;
  color: color-mix(in srgb, currentColor 80%, transparent);
}

timescrubber informative,
timescrubber timecode {
  border-radius: 7px;
  background: var(--accent-bg-color);
  color: var(--accent-fg-color);
  padding: 1px 3px;
  margin: 2px 6px;
  box-shadow: 0 2px 8px 2px rgb(0 0 0 / 27%);
  font-feature-settings: "tnum";
  font-size: .9em;
}

.utility .view {
  background: transparent;
}

.navigation-sidebar label.indicator {
  color: color-mix(in srgb, currentColor, transparent);
  border-radius: 50px;
  padding: 1px 5px;
  font-feature-settings: 'tnum';
  font-size: .8em;
}

flamegraph {
  font-size: 11px;
}

timespanlayer {
  font-size: 10px;
}

popover.tasks listview row {
  background: transparent;
  padding: 6px;
}

popover.tasks listview row label.heading {
  padding-bottom: 3px;
}

sessionfilters row {
  padding: 3px 0;
}

sessionfilters row:hover {
  background: none;
}

sessionfilters row button {
  min-width: 0;
  min-height: 0;
  padding: 3px;
}
